require 'pathname'

module Jumpup
  module Deis

    class Integrate

      attr_reader :app

      def self.deploy
        message = "Starting to deploy on Deis app #{integrate.app}"
        integrate.when_branch_send_to_deis(message) do
          integrate.deploy
        end
      end

      def self.deploy_to_production
        app = Env.all[:production_app]
        integrate = new(app)
        message = "Starting to deploy to production on Deis app #{integrate.app}"
        integrate.when_branch_send_to_deis(message) do
          integrate.deploy_to_production
        end
      end

      def self.add_remote
        message = "Adding Deis git remotes for app #{integrate.app}"
        integrate.when_branch_send_to_deis(message) do
          integrate.add_remote
        end
      end

      def self.check
        message = "Checking if there's already someone integrating to #{integrate.app}"
        integrate.when_branch_send_to_deis(message) do
          integrate.check
        end
      end

      def self.integrate
        envs = Env.all
        app = envs[:app] || envs[:staging_app]
        new(app)
      end

      def initialize(app)
        @app = app
        @envs = Env.all
      end

      def when_branch_send_to_deis(message, &block)
        puts "--> #{message}"
        if branches_that_send_to_deis.include?(actual_branch)
          block.call
        else
          puts "----> Skipping since you are in a feature branch [#{actual_branch}]"
        end
      end

      def branches_that_send_to_deis
        [Env.all[:deploy_branch], Env.all[:deploy_to_production_branch]]
      end

      def deploy
        run_database_checks if run_database_tasks?

        push(Env.all[:deploy_branch])
        migrate if run_database_tasks?
        seed if run_database_tasks?
      end

      def deploy_to_production
        run_database_checks if run_database_tasks?

        confirm_deploy
        spec
        tag
        push(Env.all[:deploy_to_production_branch])
        migrate if run_database_tasks?
        seed if run_database_tasks?
      end

      def add_remote
        remote = run_with_clean_env("git remote | grep deis").strip
        exec_with_clean_env("git remote add deis ssh://git@#{host}:2222/#{app}.git") if remote.blank?
      end

      private
        attr_reader :envs

        def check_if_migration_is_needed
          files_changed       = run_with_clean_env("git diff HEAD #{latest_remote_sha} --name-only -- db/migrate | wc -l").strip.to_i
          @migrations_changed = files_changed > 0
        end

        def check_if_seed_is_needed
          files_changed  = run_with_clean_env("git diff HEAD #{latest_remote_sha} --name-only -- db/seeds* | wc -l").strip.to_i
          @seeds_changed = files_changed > 0
        end

        def latest_remote_sha
          @latest_remote_sha ||= run_with_clean_env("git ls-remote ssh://git@#{host}:2222/#{app}.git HEAD 2>/dev/null | awk '{ print $1 }'").strip
        end

        def run_with_clean_env(command)
          Bundler.with_clean_env { `#{command}` }
        end

        def exec_with_clean_env(command)
          Bundler.with_clean_env do
            unless system(command)
              raise "Error while running #{command}"
            end
          end
        end

        def confirm(message)
          print "\n#{message}\nAre you sure? [yN] "
          raise 'Ok. Bye...' unless STDIN.gets.chomp.downcase == 'y'
        end

        def run_database_tasks?
          @envs[:run_database_tasks]
        end

        def migrate
          if @migrations_changed
            output "Migrating"
            exec_with_clean_env("deis run rake db:migrate --app #{app}")
          else
            output "Skipping migrations"
          end
        end

        def seed
          if @seeds_changed
            output "Seeding"
            exec_with_clean_env("deis run rake db:seed --app #{app}")
          else
            output "Skipping seeds"
          end
        end

        def push(branch)
          output "Pushing to #{host} from branch [#{branch}]"
          exec_with_clean_env("git push ssh://git@#{host}:2222/#{app}.git #{branch}:master")
        end

        def confirm_deploy
          confirm("[#{app}] Deploying to production using branch [#{Env.all[:deploy_to_production_branch]}]")
        end

        def spec
          puts "----> Running all specs"
          Rake::Task['spec'].invoke
        end

        def tag
          iso_date = Time.now.strftime('%Y-%m-%dT%H%M%S')
          tag_name = "production-#{iso_date}"

          puts "----> Tagging as #{tag_name}"
          exec_with_clean_env("git tag #{tag_name} master")

          puts "----> Pushing to origin"
          exec_with_clean_env("git push origin #{tag_name}")
        end

        def host
          Env.all[:host]
        end

        def actual_branch
          run_with_clean_env("git rev-parse --abbrev-ref HEAD").strip
        end

        def output(message)
          puts "----> [#{app}] #{message}"
        end

        def run_database_checks
          check_if_migration_is_needed
          check_if_seed_is_needed
        end
    end
  end
end
